<?php

abstract class AbstractQuickPayControllerCatalog extends Controller
{
    /**
     * Must be set in the extending class
     *
     * @var string
     */
    protected $module_name;

    /**
     * @var array
     */
    protected $data = array();

    /**
     * @var string
     */
    private $api_url = 'https://payment.quickpay.net/';

    /**
     * Handles the callback
     */
    public function callback()
    {
        $this->load->language('payment/' . $this->module_name);

        $callback_body = file_get_contents("php://input");
        $callback_json = json_decode($callback_body);

        $order_id = intval($callback_json->order_id);

        if ($order_id > 0) {
            $this->load->model('checkout/order');
            $transaction = end($callback_json->operations);

            if ($this->is_authorized_callback($callback_body)) {
                if ($callback_json->accepted) {
                    if ('authorize' == $transaction->type OR 'capture' == $transaction->type) {
                        $order_id = intval($order_id);

                        // Check if a fee was added to the transaction
                        if (!empty($callback_json->fee)) {
                            $this->add_transaction_fee_to_order($callback_json->fee, $order_id);
                        }
                        $this->model_checkout_order->addOrderHistory($order_id, $this->config->get($this->module_name . '_order_status_completed'));
                    }
                } else {
                    $this->log->write('The transaction for order #' . $order_id . ' was not accepted');
                }
            } else {
                $this->log->write("Order_id: $order_id. Checksum doesn't match.");
            }
        } else {
            $this->log->write('No order found!');
        }
    }

    /**
     * is_authorized_callback function.
     *
     * Performs a check on payment callbacks to see if it is legal or spoofed
     *
     * @access public
     * @return boolean
     */
    protected function is_authorized_callback($response_body)
    {
        if (!isset($_SERVER["HTTP_QUICKPAY_CHECKSUM_SHA256"])) {
            return FALSE;
        }

        return hash_hmac('sha256', $response_body, $this->config->get($this->module_name . '_privatekey')) == $_SERVER["HTTP_QUICKPAY_CHECKSUM_SHA256"];
    }

    /**
     * Adds the transaction fee to the order
     *
     * @param $raw_fee
     * @param $order_id
     */
    private function add_transaction_fee_to_order($raw_fee, $order_id)
    {
        // Fetch language
        $this->load->language('total/quickpay_fee');

        // Order Totals
        $data['totals'] = array();
        $order_total_query = $this->db->query("SELECT * FROM `" . DB_PREFIX . "order_total` WHERE order_id = '" . (int)$order_id . "' ORDER BY sort_order ASC");
        foreach ($order_total_query->rows as $total) {
            $data['totals'][] = array(
                'order_id' => $total['order_id'],
                'code' => $total['code'],
                'title' => $total['title'],
                'value' => $total['value'],
                'sort_order' => $total['sort_order']
            );
        }

        // Add fee to data object
        $fee = $raw_fee / 100;
        $data['totals'][] = array(
            'code' => 'quickpay_fee',
            'title' => $this->language->get('payment_fee'),
            'value' => floatval($fee),
            'sort_order' => 8
        );

        // Delete totals and re-add
        $this->db->query("DELETE FROM " . DB_PREFIX . "order_total WHERE order_id = '" . (int)$order_id . "'");

        if (isset($data['totals'])) {
            foreach ($data['totals'] as $total) {
                // Add fee to the total price
                if ($total['code'] === 'total') {
                    $total['value'] += $fee;

                    // Update the order entry
                    $this->db->query("UPDATE " . DB_PREFIX . "order SET total = '{$total['value']}' WHERE order_id = '{$order_id}'");

                }
                // Add totals to the DB
                $this->db->query("INSERT INTO " . DB_PREFIX . "order_total SET order_id = '" . (int)$order_id . "', code = '" . $this->db->escape($total['code']) . "', title = '" . $this->db->escape($total['title']) . "', `value` = '" . (float)$total['value'] . "', sort_order = '" . (int)$total['sort_order'] . "'");
            }
        }

        // Add notice in the log
        $this->log->write('Add fee: ' . $fee . ' to order: ' . $order_id);
    }

    /**
     * Instantiation logic on controller construction
     */
    protected function init()
    {
        $this->language->load('payment/' . $this->module_name);

        $this->load->model('checkout/order');

        $this->populate_language_data();

        $this->populate_config_data();

        $this->populate_common_data();

        if (version_compare(VERSION, "2.2.0.0") >= 0) {
            return $this->load->view('payment/' . $this->module_name .'.tpl', $this->data);
        } else {
            if (file_exists(DIR_TEMPLATE . $this->config->get('config_template') . '/template/payment/'. $this->module_name .'.tpl')) {
                return $this->load->view( $this->config->get('config_template') . '/template/payment/'. $this->module_name .'.tpl', $this->data);
            } else {
                return $this->load->view('default/template/payment/'. $this->module_name .'.tpl', $this->data);
            }
        }
    }

    /**
     * Populates a data array.
     */
    protected function populate_language_data()
    {
        // Language fields
        $fields = array(
            'text_credit_card',
            'text_start_date',
            'text_issue',
            'text_wait',
            'button_confirm',
            'button_back'
        );

        // Retrieve fields which is unique for the instance
        $unique_fields = $this->get_unique_language_data();

        // Merge the fields together
        $fields = array_merge($fields, $unique_fields);

        // Populate language fields with data
        foreach ($fields as $field) {
            $this->data[$field] = $this->language->get($field);
        }
    }

    /**
     * Returns language fields unique for a gateway
     * @return array
     */
    abstract protected function get_unique_language_data();

    /**
     * Populates the config data
     */
    protected function populate_config_data()
    {
        // Config fields
        $fields = array(
            'merchant',
            'agreement',
            'currency',
            'order_status_id',
            'autocapture',
            'autofee',
            'text_on_statement',
            'branding_id'
        );

        // Populate config fields
        foreach ($fields as $field) {
            $field = $this->module_name . '_' . $field;
            $this->data[$field] = $this->config->get($field);
        }
    }

    /**
     * Populates common data
     */
    protected function populate_common_data()
    {
        $this->data['action'] = $this->api_url;

        $this->data[$this->module_name . '_language'] = $this->language->get('code');

        $this->data[$this->module_name . '_continueurl'] = $this->url->link('checkout/success', '', 'SSL');
        $this->data[$this->module_name . '_cancelurl'] = $this->url->link('checkout/checkout', '', 'SSL');
        $this->data[$this->module_name . '_callbackurl'] = $this->url->link('payment/' . $this->module_name . '/callback', '', 'SSL');
        $this->data[$this->module_name . '_version'] = 'v10';

        $this->data['back'] = $this->data[$this->module_name . '_cancelurl'];

        $order_info = $this->model_checkout_order->getOrder($this->session->data['order_id']);

        $this->data['order_id'] = str_pad($this->session->data['order_id'], 4, '0', STR_PAD_LEFT);
        $this->data['amount'] = 100 * $this->currency->format($order_info['total'], $order_info['currency_code'], '', FALSE);
        $this->data['currency'] = $order_info['currency_code'];

        $this->data['quickpay_cardtypelock'] = $this->get_cardtype_lock();

        $this->data['quickpay_basket'] = $this->get_basket_array();
        $this->data['quickpay_billing'] = $this->get_billing_array($order_info);
        $this->data['quickpay_shipping'] = $this->data['quickpay_billing'];

        $this->data[$this->module_name . '_customer_email'] = $this->get_customer_email();

        $this->data['quickpay_checksum'] = $this->create_md5();
    }

    /**
     * Returns the cardtype lock for a payment gateway
     * @return string
     */
    abstract protected function get_cardtype_lock();

    /**
     * Creates a basket form data array
     *
     * @return array
     */
    protected function get_basket_array()
    {
        $basket = array();

        foreach ($this->cart->getProducts() as $product) {
            $basket[] = array(
                'qty' => $product['quantity'],
                'item_no' => $product['product_id'],
                'item_name' => html_entity_decode($product['name']),
                'item_price' => $this->tax->calculate($product['price'], $product['tax_class_id']) * 100,
                'vat_rate' => $this->calculate_basket_item_taxes_rate($product['price'], $product['tax_class_id'])
            );
        }

        return $basket;
    }

    /**
     * Calculates a single basket item tax rate
     *
     * @param $price
     * @param $tax_class_id
     * @return float|int
     */
    protected function calculate_basket_item_taxes_rate($price, $tax_class_id)
    {
        $taxes = $this->tax->getRates($price, $tax_class_id);
        $rate = 0;

        if (!empty($taxes) && is_array($taxes)) {
            foreach ($taxes as $tax) {
                $rate += floatval($tax['rate']);
            }
        }
        return $rate > 0 ? ($rate / 100) : $rate;
    }

    /**
     * Creates a billing form data array
     *
     * @param $order_info
     * @return array
     */
    protected function get_billing_array($order_info)
    {
        return array(
            'name' => html_entity_decode($order_info['payment_firstname'] . ' ' . $order_info['payment_lastname'], ENT_QUOTES, 'UTF-8'),
            'street' => html_entity_decode($order_info['payment_address_1'], ENT_QUOTES, 'UTF-8'),
            'zip_code' => html_entity_decode($order_info['payment_postcode'], ENT_QUOTES, 'UTF-8'),
            'city' => html_entity_decode($order_info['payment_city'], ENT_QUOTES, 'UTF-8'),
            'country_code' => html_entity_decode($order_info['payment_iso_code_3'], ENT_QUOTES, 'UTF-8'),
            'phone_number' => html_entity_decode($order_info['telephone'], ENT_QUOTES, 'UTF-8'),
            'email' => $order_info['email'],
        );
    }

    /**
     * Returns the customer name
     * @return string
     */
    protected function get_customer_email()
    {
        if (isset($this->session->data['guest'])) {
            $customer_email = $this->session->data['guest']['email'];
        } else {
            $customer_email = $this->customer->getEmail();
        }
        return $customer_email;
    }

    /**
     * Returns a MD5 checksum based on the form data
     * @return string
     */
    protected function create_md5()
    {
        $params = array(
            'agreement_id' => $this->data[$this->module_name . '_agreement'],
            'merchant_id' => $this->data[$this->module_name . '_merchant'],
            'language' => $this->data[$this->module_name . '_language'],
            'order_id' => $this->data['order_id'],
            'amount' => $this->data['amount'],
            'currency' => $this->data['currency'],
            'continueurl' => $this->data[$this->module_name . '_continueurl'],
            'cancelurl' => $this->data[$this->module_name . '_cancelurl'],
            'callbackurl' => $this->data[$this->module_name . '_callbackurl'],
            'autocapture' => $this->data[$this->module_name . '_autocapture'],
            'autofee' => $this->data[$this->module_name . '_autofee'],
            'payment_methods' => $this->data['quickpay_cardtypelock'],
            'customer_email' => $this->data[$this->module_name . '_customer_email'],
            'version' => $this->data[$this->module_name . '_version'],
            'text_on_statement' => $this->data[$this->module_name . '_text_on_statement'],
            'branding_id' => $this->data[$this->module_name . '_branding_id'],
        );

        if (!empty($this->data['quickpay_billing'])) {
            foreach ($this->data['quickpay_billing'] as $index => $value) {
                $params["invoice_address[$index]"] = $value;
            }
        }

        if (!empty($this->data['quickpay_shipping'])) {
            foreach ($this->data['quickpay_billing'] as $index => $value) {
                $params["shipping_address[$index]"] = $value;
            }
        }

        if (!empty($this->data['quickpay_basket'])) {
            foreach ($this->data['quickpay_basket'] as $basket_line_number => $basket_lines) {
                foreach ($basket_lines as $basket_item_key => $basket_item) {
                    $params["basket[" . $basket_line_number . "][" . $basket_item_key . "]"] = $basket_item;
                }
            }
        }

        ksort($params);

        $checksum = hash_hmac("sha256", implode(" ", $params), $this->config->get($this->module_name . '_apikey'));

        return $checksum;
    }

    /**
     * Convenience wrapper to get config data
     *
     * @param $field
     * @param string $default
     * @return string
     */
    protected function get_setting($field, $default = '')
    {
        $data = $this->config->get($field);

        if (!empty($data)) {
            return $data;
        }

        return $default;
    }
}